/*
---------------------------------------------------------------------------
|					File functions/serialization
|				  Copyright (C) 1998 VF Technologies 
|	 					Code: Rik Arends
|						All Rights Reserved
---------------------------------------------------------------------------
*/

//---------------------------------------------------------------------------
DECLARE_CLASS(SerialIO);
class  SerialIO // virtual serial IO class.. used to serialize stuff (maybe even over network)
{
public:

	SerialIO()				{cur_position=0,Mode=mode_readwrite;Opened=false;}
	virtual ~SerialIO()		{};

	enum
	{
		mode_read		=1,
		mode_write		=2,
		mode_readwrite	=mode_read|mode_write,  // normal access mode
		mode_streamread	=mode_read |4,	 // network transferring modes.. no seeking possible, only forward skipping
		mode_streamwrite=mode_write|4, 
	};

	virtual bool Write(const void *buf, long len)=0;					// write
	virtual bool WriteAt(long offset,const void *buf, long len)=0;		// write at offset (no filepointer change)
	virtual bool Read(void *buf,long len)=0;
	virtual bool ReadAt(long offset,void *buf,long len)=0;

	virtual bool	IsEnd()=0;												// Check for end-of-block
	virtual long	GetPos()=0;												// Return offset in block (file, mem or network)

	virtual void	Seek(long new_pos)=0;									// Seek from beginning of file
	virtual long	Skip(long amount)=0;									// Seek from current position
	virtual void	Flush()=0;												// Flush (write) buffer if applicable
	// Attributes
	virtual long	GetLength()=0;											// Get length of file
	inline cString	GetError() const										{ return Error; }
	inline void		ClearError()											{ Error.Clear(); }

	// Implementation
	bool	Opened;													// is open?
	long	cur_position;											// current position in block
	String	Error;													// Error message
	short	Mode;													// mode (read/write/stream etc..)
};

//---------------------------------------------------------------------------

DECLARE_CLASS(FileIO);
class  FileIO:public SerialIO
{
public:

	enum	{	FILEBUFFERSIZE = 4096	};

	FileIO()				{file=NULL;}
	virtual ~FileIO()		{Close();} //NOTE.. TODO.. gives a weird call to pure virtual function on delete.. CHek this OUT!

	bool Open(cString filename,long mode);
	bool Close(); 

	virtual bool Write(const void *buf, long len);						// write
	virtual bool WriteAt(long offset,const void *buf, long len);		// write at offset NOT BUFFERED
	virtual bool Read(void *buf,long len);								// reads bytes..
	virtual bool ReadAt(long offset,void *buf,long len);				// Not buffered

	// some macro functions
	void	Write_byte(byte b)				{Write(&b,1);}
	void	Write_word(word w)				{Write(&w,2);}		
	void	Write_dword(dword d)			{Write(&d,4);}
	void	Write_qword(qword q)			{Write(&q,8);}
	byte	Read_byte()						{byte  b;Read(&b,1);return b;}
	word	Read_word()						{word  w;Read(&w,2);return w;}
	dword	Read_dword()					{dword d;Read(&d,4);return d;}
	qword	Read_qword()					{qword q;Read(&q,8);return q;}

	virtual bool	IsEnd();											// Check for end-of-block
	virtual long	GetPos();											// Return offset in block (including buffer)

	virtual void	Seek(long new_pos);									// Seek from beginning of file
	virtual long	Skip(long amount);									// Seek from current position
	virtual void	Flush();											// Flush (write) buffer if applicable

	// Attributes
	long GetLength();													// Get length of file

	// privat methods

	// file buffer..
	long	GetFilePos();										// Return offset in file (ex buffer)
	void    FlushBuffer();										// flushes write buffer
	void    ReadBuffer();										// Reads next block of BUFFERSIZE in buffer

	struct _buffer // contains file buffer info..
	{
		char 	mem[FILEBUFFERSIZE];
		long	size;							 // Size of the buffer
		long	used;							 // number of bytes used
		long	start;						     // offset in file the buffer starts..
		long    readoffset;						 // read offset in buffer
	}buf;

	long	filesize;							 // size of file..
	void    *file;								 // filehandle (!=NULL if opened)
};

//---------------------------------------------------------------------------

DECLARE_CLASS(Archive);
class  Archive
{
public:
	// Constants/defines
	enum { Save, Load };

	Archive()		{Opened=false;target=NULL;HeaderOffset=0;}
	~Archive()		{Close();}

	bool	Open(SerialIO *i_target, short type,long version, long ID=0, bool writeheader=true); // save and load determines what is done.. read or write
	bool	OpenOld(SerialIO *i_target, short type,long version, bool writeheader=true); // save and load determines what is done.. read or write
	void	Close(); // writes file size to header (if applicable)

	// I/O
	void	JumpTo(long offset);		// skips to offset(FORWARD ONLY)

	long    OffsetTagGet();			      // Just reads the offset tag
	void    OffsetTagJump(long offset=-1);// If offset passed, jumps there.. else reads offset tag and jumps
	long	OffsetTagPlace();			  // writes an offset tag, returns offset for passing to offsettagSet
	void	OffsetTagSet(long offset);    // Writes current offset to tag at 'offset'

	void	Write_byte(byte b);
	void	Write_word(word w);
	void	Write_dword(dword d);
	void	Write_qword(qword q);
	byte	Read_byte();
	word	Read_word();
	dword	Read_dword();
	qword	Read_qword();

	void	Write(const void *buf, uint len);
	bool	Read(void *buf, uint len);

	enum // ID's for recognizing objects in the fileformat (debugging mostly)
	{
		FID_CHAR	=10,
		FID_INT		=11,
		FID_SHORT	=12,
		FID_LONG	=13,
		FID_FLOAT	=14,
		FID_DOUBLE	=15,
		FID_BOOL	=16,
		FID_STRING	=17,
		FID_BLOCK	=18,
		FID_BYTE	=19,
		FID_WORD	=20,
		FID_DWORD	=21,
		FID_QWORD	=22
	};

	friend  rArchive operator<<(rArchive arch, const char   d)		{arch.Write_byte(FID_CHAR);  arch.Write_byte( *((byte  *)&d)); return arch;};
	friend  rArchive operator<<(rArchive arch, const int    d)		{arch.Write_byte(FID_INT);   arch.Write_dword(*((dword *)&d)); return arch;};	
	friend  rArchive operator<<(rArchive arch, const short  d)		{arch.Write_byte(FID_SHORT); arch.Write_word( *((word  *)&d)); return arch;};		
	friend  rArchive operator<<(rArchive arch, const long   d)		{arch.Write_byte(FID_LONG);  arch.Write_dword(*((dword *)&d)); return arch;};
	friend  rArchive operator<<(rArchive arch, const float  d)		{arch.Write_byte(FID_FLOAT); arch.Write_dword(*((dword *)&d)); return arch;};
	friend  rArchive operator<<(rArchive arch, const double d)		{arch.Write_byte(FID_DOUBLE);arch.Write_qword(*((qword *)&d)); return arch;};	
	friend  rArchive operator<<(rArchive arch, const bool   d)		{arch.Write_byte(FID_BOOL);  arch.Write_byte (*((byte  *)&d)); return arch;};	
	friend  rArchive operator<<(rArchive arch, const String &d)		{arch.Write_byte(FID_STRING);arch.Write_word(d.Length());arch.Write(d.GetReadPtr(),d.Length()); return arch;};	

	// simple types
	friend  rArchive operator<<(rArchive arch, const byte   &d)		{arch.Write_byte(FID_BYTE);  arch.Write_byte( d); return arch;};
	friend  rArchive operator<<(rArchive arch, const word   &d)		{arch.Write_byte(FID_WORD);  arch.Write_word(d);  return arch;};	
	friend  rArchive operator<<(rArchive arch, const dword  &d)		{arch.Write_byte(FID_DWORD); arch.Write_dword( d);return arch;};		
	friend  rArchive operator<<(rArchive arch, const qword  &d)		{arch.Write_byte(FID_QWORD); arch.Write_qword(d); return arch;};

	// normal types
	friend  rArchive operator>>(rArchive arch, char   &d)		{byte x=arch.Read_byte();assert(x==FID_CHAR);  byte  c=arch.Read_byte(); d=*((char   *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, int    &d)		{byte x=arch.Read_byte();assert(x==FID_INT);   dword c=arch.Read_dword();d=*((int    *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, short  &d)		{byte x=arch.Read_byte();assert(x==FID_SHORT); word  c=arch.Read_word(); d=*((short  *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, long   &d)		{byte x=arch.Read_byte();assert(x==FID_LONG);  dword c=arch.Read_dword();d=*((long   *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, float  &d)		{byte x=arch.Read_byte();assert(x==FID_FLOAT); dword c=arch.Read_dword();d=*((float  *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, double &d)		{byte x=arch.Read_byte();assert(x==FID_DOUBLE);qword c=arch.Read_qword();d=*((double *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, bool   &d)		{byte x=arch.Read_byte();assert(x==FID_BOOL);  byte  c=arch.Read_byte(); d=*((bool   *)&c); return arch;};
	friend  rArchive operator>>(rArchive arch, String &d)		{byte x=arch.Read_byte();assert(x==FID_STRING);long len=arch.Read_word();d.SetLength(len);arch.Read(d.GetPtr(),len); return arch;};	

	// simple types
	friend  rArchive operator>>(rArchive arch, byte   &d)		{byte x=arch.Read_byte();assert(x==FID_BYTE);   d=arch.Read_byte(); return arch;};
	friend  rArchive operator>>(rArchive arch, word   &d)		{byte x=arch.Read_byte();assert(x==FID_WORD);   d=arch.Read_word(); return arch;};
	friend  rArchive operator>>(rArchive arch, dword  &d)		{byte x=arch.Read_byte();assert(x==FID_DWORD);  d=arch.Read_dword();return arch;};
	friend  rArchive operator>>(rArchive arch, qword  &d)		{byte x=arch.Read_byte();assert(x==FID_QWORD);  d=arch.Read_qword();return arch;};

	// Status
	dword	GetPos() const;									
	float	Progress() const;								// in load mode returns fraction between pos and size
	long	Version() const									{ return ArcHdr.Version; }

	// Error management
	inline cString	GetError() const						{ return Error; }
	inline void		ClearError()							{ Error.Clear(); }

protected:
	// Archive management
	bool    HasHeader;
	bool	Opened;													// Opened archive?
	long	type;

	SerialIO *target;												// -> Current associated serial IO thing
	String	Error;													// Error code

	long	HeaderOffset;											// the position in the output stream when the Open() was called
	struct
	{
		char	Tag[4];												// File ID tag = VFAR
		long	Version;											// Version number
		long	ID;													// file sub-ID
		long	FileLength;											// File length & CRC
	} ArcHdr;
};
